#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "strm.h"

#define STREAMTAGSZ sizeof(struct StreamTag)
struct StreamTag
{
  int size;
  int offset;
  void * data;
  const char * name;
};

static iff_size dear(struct StreamTag *, void *, iff_size);
static iff_size seek(struct StreamTag *, iff_size, int);
static iff_size tell(struct StreamTag *);
static iff_void done(struct StreamTag *);

typedef iff_size (* _iread)(void *, void *, iff_size);
typedef iff_size (* _iseek)(void *, iff_size, int);
typedef iff_size (* _itell)(void *);
typedef iff_void (* _idone)(void *);

struct IFFStreamTag * stream_init(const char * name)
{
  FILE * fp;
  struct StreamTag * stream;
  struct IFFStreamTag * iff_stream;
  long read;

  stream = malloc(STREAMTAGSZ);
  iff_stream = malloc(IFFSTREAMTAGSZ);

  _CHECK(!stream,     "malloc(%u) failed", STREAMTAGSZ, 0);
  _CHECK(!iff_stream, "malloc(%u) failed", IFFSTREAMTAGSZ, 0);

  stream->size            = 0;
  stream->offset          = 0;
  stream->data            = 0;
  stream->name            = name;
  iff_stream->stream      = stream;
  iff_stream->stream_name = name; /* sucks */

  iff_stream->stream_read = (_iread) &dear;
  iff_stream->stream_seek = (_iseek) &seek;
  iff_stream->stream_tell = (_itell) &tell;
  iff_stream->stream_done = (_idone) &done;

  fp = fopen(name, "rb");
  _CHECK(!fp, "can't open %s", name, 0);

  fseek(fp, 0, SEEK_END);
  stream->size = ftell(fp);
  _CHECK(!stream->size, "%s is empty", name, 0);

  fseek(fp, 0, SEEK_SET);
  stream->data = malloc(stream->size);

  _CHECK(!stream->data, "malloc(%u) failed", stream->size, 0);

  read = fread(stream->data, stream->size, 1, fp);
  _CHECK(read != 1, "failed to read %u bytes from %s", stream->size, name);

  fclose(fp);

  /*
   * for alignment to work
   */
  stream->size = (stream->size + 1) & ~1;

  return iff_stream;
}

static iff_size dear(struct StreamTag * stream, void * data, iff_size size)
{
  _CHECK(stream->offset + size > stream->size, 
         "attempt to read past the end of stream %u %s",
         stream->offset + size, stream->name);
  memcpy(data, (char *) stream->data + stream->offset, size);
  stream->offset += size;
  return size;
}

static iff_size seek(struct StreamTag * stream, iff_size size, int kind)
{
  iff_size offset = 0xffffffff;

  switch(kind)
  {
    case 0: offset = size; break;
    case 1: offset = stream->offset + size; break;
    case 2: offset = stream->size - size; break;
    default:
      _CHECK(kind, "bogus %d seek(%u) method called", kind, size);
      break;
  }
  _CHECK(offset > stream->size, 
         "attempt to seek past the end of stream %u %s",
         offset, stream->name);

  stream->offset = offset;
  return size;
}

static iff_size tell(struct StreamTag * stream)
{
  return stream->offset;
}

static iff_void done(struct StreamTag * stream)
{
  free(stream->data);
  free(stream);
}
