#include "config.h"

#include "mtx.h"

/* row major (i hope) */

void mtx_identity(float * r)
{
  static float identity[9] =
  {
    1.f, 0.f, 0.f,
    0.f, 1.f, 0.f,
    0.f, 0.f, 1.f
  };
  MTXCPY(r, identity);
}

void mtx_rotationcs(float * r, float c, float s, float x, float y, float z) 
{
  float xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c; 
  float mag = (float) sqrt(x * x + y * y + z * z);
  if (mag == 0.f)
  {
    mtx_identity(r);
    return;
  }
  x /= mag;
  y /= mag;
  z /= mag;

  xx = x * x;
  yy = y * y;
  zz = z * z;
  xy = x * y;
  yz = y * z;
  zx = z * x;
  xs = x * s;
  ys = y * s;
  zs = z * s;
  one_c = 1.f - c;

  r[0] = (one_c * xx) + c;
  r[3] = (one_c * xy) - zs;
  r[6] = (one_c * zx) + ys;

  r[1] = (one_c * xy) + zs;
  r[4] = (one_c * yy) + c;
  r[7] = (one_c * yz) - xs;

  r[2] = (one_c * zx) - ys;
  r[5] = (one_c * yz) + xs;
  r[8] = (one_c * zz) + c;
}

void mtx_rotation(float * r, float angle, float x, float y, float z)
{
  mtx_rotationcs(r, (float) cos(angle * DEG2RAD), (float) sin(angle * DEG2RAD),
                 x, y, z);
}

void mtx_mtxmul(float * r, float * a, float * b)
{
  int i;

  for (i = 0; i < 3; i++)
  {
    #define I + i * 3
    r[0 I] = a[0 I] * b[0] + a[1 I] * b[3] + a[2 I] * b[6];
    r[1 I] = a[0 I] * b[1] + a[1 I] * b[4] + a[2 I] * b[7];
    r[2 I] = a[0 I] * b[2] + a[1 I] * b[5] + a[2 I] * b[8];
    #undef I
  }
}

void mtx_vecmul(float * r, float * m, float x, float y, float z)
{
  r[0] = x * m[0] + y * m[3] + z * m[6];
  r[1] = x * m[1] + y * m[4] + z * m[7];
  r[2] = x * m[2] + y * m[5] + z * m[8];
}

#include <stdio.h>
void mtx_printf(float * m)
{
  printf("%9f %9f %9f\n", m[0], m[1], m[2]);
  printf("%9f %9f %9f\n", m[3], m[4], m[5]);
  printf("%9f %9f %9f\n", m[6], m[7], m[8]);
  fflush(stdout);
}

#ifdef _TEST_
int main(void)
{
}
#endif
